@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2016-2023 Advanced Micro Devices, Inc. All rights reserved.
+ * Copyright (C) 2016-2024 Advanced Micro Devices, Inc. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -23,19 +23,18 @@
 
 #pragma once
 
-#include "hipblas.h"
+#ifdef _OPENMP
+#include <omp.h>
+#endif
+#include <hipblas/hipblas.h>
 #include <stdbool.h>
 
+#include "type_utils.h"
 #ifdef __cplusplus
-#include "cblas_interface.h"
-#include "complex.hpp"
 #include "hipblas_datatype2string.hpp"
-#include <cmath>
 #include <cstdio>
-#include <immintrin.h>
 #include <iostream>
 #include <random>
-#include <type_traits>
 #include <vector>
 #endif
 
@@ -49,42 +48,8 @@
  * \brief provide data initialization, timing, hipblas type <-> lapack char conversion utilities.
  */
 
-#define CHECK_HIP_ERROR(error)                        \
-    do                                                \
-    {                                                 \
-        hipError_t error__ = (error);                 \
-        if(error__ != hipSuccess)                     \
-        {                                             \
-            fprintf(stderr,                           \
-                    "hip error: '%s'(%d) at %s:%d\n", \
-                    hipGetErrorString(error__),       \
-                    error__,                          \
-                    __FILE__,                         \
-                    __LINE__);                        \
-            exit(EXIT_FAILURE);                       \
-        }                                             \
-    } while(0)
-
 #ifdef __cplusplus
 
-#ifndef CHECK_HIPBLAS_ERROR
-#define EXPECT_HIPBLAS_STATUS(status, expected)      \
-    do                                               \
-    {                                                \
-        hipblasStatus_t status__ = (status);         \
-        if(status__ != expected)                     \
-        {                                            \
-            fprintf(stderr,                          \
-                    "hipBLAS error: %s at %s:%d\n",  \
-                    hipblasStatusToString(status__), \
-                    __FILE__,                        \
-                    __LINE__);                       \
-            return (status__);                       \
-        }                                            \
-    } while(0)
-#define CHECK_HIPBLAS_ERROR(STATUS) EXPECT_HIPBLAS_STATUS(STATUS, HIPBLAS_STATUS_SUCCESS)
-#endif
-
 #define BLAS_1_RESULT_PRINT                                \
     do                                                     \
     {                                                      \
@@ -106,100 +71,28 @@
         }                                                  \
     } while(0)
 
-// Return true if value is NaN
-template <typename T>
-inline bool hipblas_isnan(T)
-{
-    return false;
-}
-inline bool hipblas_isnan(double arg)
-{
-    return std::isnan(arg);
-}
-inline bool hipblas_isnan(float arg)
-{
-    return std::isnan(arg);
-}
-
-inline bool hipblas_isnan(hipblasHalf arg)
-{
-    auto half_data = static_cast<unsigned short>(arg);
-    return (~(half_data)&0x7c00) == 0 && (half_data & 0x3ff) != 0;
-}
-inline bool hipblas_isnan(hipblasComplex arg)
-{
-    return std::isnan(arg.real()) || std::isnan(arg.imag());
-}
-inline bool hipblas_isnan(hipblasDoubleComplex arg)
-{
-    return std::isnan(arg.real()) || std::isnan(arg.imag());
-}
-
-inline hipblasHalf float_to_half(float val)
-{
-#ifdef HIPBLAS_USE_HIP_HALF
-    return __float2half(val);
-#else
-    uint16_t a = _cvtss_sh(val, 0);
-    return a;
-#endif
-}
-
-inline hipblasHalf float_to_half(hipblasBfloat16 val)
-{
-    return float_to_half(float(val));
-}
-
-inline float half_to_float(hipblasHalf val)
-{
-#ifdef HIPBLAS_USE_HIP_HALF
-    return __half2float(val);
-#else
-    return _cvtsh_ss(val);
-#endif
-}
-
-inline std::ostream& operator<<(std::ostream& os, const hipblasBfloat16& bf)
-{
-    return os << float(bf);
-}
-
-inline float bfloat16_to_float(hipblasBfloat16 bf)
-{
-    return hipblasBfloat16::bfloat16_to_float(bf);
-}
-
-inline hipblasBfloat16 float_to_bfloat16(float f)
-{
-    return hipblasBfloat16::float_to_bfloat16(f);
-}
-
 /* =============================================================================================== */
-/* Complex / real helpers.                                                                         */
-template <typename T>
-static constexpr bool is_complex = false;
-
+/* Epsilon helpers for near checks.                                                                */
+template <typename>
+HIPBLAS_CLANG_STATIC constexpr double hipblas_type_epsilon = 0;
 template <>
-HIPBLAS_CLANG_STATIC constexpr bool is_complex<hipblasComplex> = true;
-
+HIPBLAS_CLANG_STATIC constexpr double
+    hipblas_type_epsilon<float> = std::numeric_limits<float>::epsilon();
 template <>
-HIPBLAS_CLANG_STATIC constexpr bool is_complex<hipblasDoubleComplex> = true;
-
-// Get base types from complex types.
-template <typename T, typename = void>
-struct real_t_impl
-{
-    using type = T;
-};
-
-template <typename T>
-struct real_t_impl<T, std::enable_if_t<is_complex<T>>>
-{
-    using type = decltype(T{}.real());
-};
-
-template <typename T>
-using real_t = typename real_t_impl<T>::type;
+HIPBLAS_CLANG_STATIC constexpr double
+    hipblas_type_epsilon<double> = std::numeric_limits<double>::epsilon();
+template <>
+HIPBLAS_CLANG_STATIC constexpr double
+    hipblas_type_epsilon<hipblasComplex> = std::numeric_limits<float>::epsilon();
+template <>
+HIPBLAS_CLANG_STATIC constexpr double
+    hipblas_type_epsilon<hipblasDoubleComplex> = std::numeric_limits<double>::epsilon();
+template <>
+HIPBLAS_CLANG_STATIC constexpr double hipblas_type_epsilon<
+    hipblasHalf> = 0.0009765625; // in fp16 diff between 0x3C00 (1.0) and fp16 0x3C01
+template <>
+HIPBLAS_CLANG_STATIC constexpr double hipblas_type_epsilon<
+    hipblasBfloat16> = 0.0078125; // in bf16 diff between 0x3F80 (1.0) and bf16 0x3F81 in double precision
 
 /* ============================================================================================ */
 /*! \brief  Random number generator which generates NaN values */
@@ -360,7 +253,6 @@ inline hipblasBfloat16 random_generator<hipblasBfloat16>()
 template <>
 inline hipblasComplex random_generator<hipblasComplex>()
 {
-    return hipblasComplex(rand() % 10 + 1, rand() % 10 + 1);
     return {float(rand() % 10 + 1), float(rand() % 10 + 1)};
 }
 
@@ -369,7 +261,6 @@ inline hipblasComplex random_generator<hipblasComplex>()
 template <>
 inline hipblasDoubleComplex random_generator<hipblasDoubleComplex>()
 {
-    return hipblasDoubleComplex(rand() % 10 + 1, rand() % 10 + 1);
     return {double(rand() % 10 + 1), double(rand() % 10 + 1)};
 }
 
@@ -427,7 +318,14 @@ template <>
 inline hipblasBfloat16 random_hpl_generator()
 {
     return hipblasBfloat16(
-        float_to_bfloat16(std::uniform_real_distribution<float>(-0.5, 0.5)(hipblas_rng)));
+        float_to_bfloat16(std::uniform_real_distribution<float>(-0.5f, 0.5f)(hipblas_rng)));
+}
+
+template <>
+inline hipblasHalf random_hpl_generator()
+{
+    return hipblasHalf(
+        float_to_half(std::uniform_real_distribution<float>(-0.5f, 0.5f)(hipblas_rng)));
 }
 
 /* ============================================================================================ */
@@ -436,27 +334,32 @@ inline hipblasBfloat16 random_hpl_generator()
 /*! \brief  matrix/vector initialization: */
 // for vector x (M=1, N=lengthX, lda=incx);
 // for complex number, the real/imag part would be initialized with the same value
+template <typename T>
+void hipblas_init(std::vector<T>& A,
+                  int64_t         M,
+                  int64_t         N,
+                  int64_t         lda,
+                  hipblasStride   stride      = 0,
+                  int64_t         batch_count = 1)
+{
+    for(int64_t b = 0; b < batch_count; b++)
+        for(int64_t i = 0; i < M; ++i)
+            for(int64_t j = 0; j < N; ++j)
+                A[i + j * lda + b * stride] = random_generator<T>();
+}
+
 template <typename T>
 void hipblas_init(
-    std::vector<T>& A, int M, int N, int lda, hipblasStride stride = 0, int batch_count = 1)
+    T* A, int64_t M, int64_t N, int64_t lda, hipblasStride stride = 0, int64_t batch_count = 1)
 {
-    for(int b = 0; b < batch_count; b++)
-        for(int i = 0; i < M; ++i)
-            for(int j = 0; j < N; ++j)
+    for(int64_t b = 0; b < batch_count; b++)
+        for(int64_t i = 0; i < M; ++i)
+            for(int64_t j = 0; j < N; ++j)
                 A[i + j * lda + b * stride] = random_generator<T>();
 }
 
 template <typename T>
-void hipblas_init(T* A, int M, int N, int lda, hipblasStride stride = 0, int batch_count = 1)
-{
-    for(int b = 0; b < batch_count; b++)
-        for(int i = 0; i < M; ++i)
-            for(int j = 0; j < N; ++j)
-                A[i + j * lda + b * stride] = random_generator<T>();
-}
-
-template <typename T>
-void hipblas_init_alternating_sign(std::vector<T>& A, int M, int N, int lda)
+void hipblas_init_alternating_sign(std::vector<T>& A, int64_t M, int64_t N, int64_t lda)
 {
     // Initialize matrix so adjacent entries have alternating sign.
     // In gemm if either A or B are initialized with alernating
@@ -465,8 +368,8 @@ void hipblas_init_alternating_sign(std::vector<T>& A, int M, int N, int lda)
     // This helps reduce floating point inaccuracies for 16bit
     // arithmetic where the exponent has only 5 bits, and the
     // mantissa 10 bits.
-    for(int i = 0; i < M; ++i)
-        for(int j = 0; j < N; ++j)
+    for(int64_t i = 0; i < M; ++i)
+        for(int64_t j = 0; j < N; ++j)
             if(j % 2 ^ i % 2)
                 A[i + j * lda] = random_generator<T>();
             else
@@ -475,7 +378,7 @@ void hipblas_init_alternating_sign(std::vector<T>& A, int M, int N, int lda)
 
 template <typename T>
 void hipblas_init_alternating_sign(
-    std::vector<T>& A, int M, int N, int lda, hipblasStride stride, int batch_count)
+    std::vector<T>& A, int64_t M, int64_t N, int64_t lda, hipblasStride stride, int64_t batch_count)
 {
     // Initialize matrix so adjacent entries have alternating sign.
     // In gemm if either A or B are initialized with alernating
@@ -484,9 +387,9 @@ void hipblas_init_alternating_sign(
     // This helps reduce floating point inaccuracies for 16bit
     // arithmetic where the exponent has only 5 bits, and the
     // mantissa 10 bits.
-    for(int i_batch = 0; i_batch < batch_count; i_batch++)
-        for(int i = 0; i < M; ++i)
-            for(int j = 0; j < N; ++j)
+    for(int64_t i_batch = 0; i_batch < batch_count; i_batch++)
+        for(int64_t i = 0; i < M; ++i)
+            for(int64_t j = 0; j < N; ++j)
                 if(j % 2 ^ i % 2)
                     A[i + j * lda + i_batch * stride] = random_generator<T>();
                 else
@@ -499,7 +402,9 @@ void hipblas_init_hpl_alternating_sign(
     T* A, size_t M, size_t N, size_t lda, size_t stride = 0, size_t batch_count = 1)
 {
     for(size_t i_batch = 0; i_batch < batch_count; i_batch++)
+#ifdef _OPENMP
 #pragma omp parallel for
+#endif
         for(size_t j = 0; j < N; ++j)
         {
             size_t offset = j * lda + i_batch * stride;
@@ -544,7 +449,9 @@ inline void
     hipblas_init_cos(T* A, size_t M, size_t N, size_t lda, size_t stride, size_t batch_count)
 {
     for(size_t i_batch = 0; i_batch < batch_count; i_batch++)
+#ifdef _OPENMP
 #pragma omp parallel for
+#endif
         for(size_t j = 0; j < N; ++j)
         {
             size_t offset = j * lda + i_batch * stride;
@@ -565,7 +472,9 @@ inline void
     hipblas_init_sin(T* A, size_t M, size_t N, size_t lda, size_t stride, size_t batch_count)
 {
     for(size_t i_batch = 0; i_batch < batch_count; i_batch++)
+#ifdef _OPENMP
 #pragma omp parallel for
+#endif
         for(size_t j = 0; j < N; ++j)
         {
             size_t offset = j * lda + i_batch * stride;
@@ -586,7 +495,9 @@ inline void hipblas_init_cos<hipblasBfloat16>(
     hipblasBfloat16* A, size_t M, size_t N, size_t lda, size_t stride, size_t batch_count)
 {
     for(size_t i_batch = 0; i_batch < batch_count; i_batch++)
+#ifdef _OPENMP
 #pragma omp parallel for
+#endif
         for(size_t j = 0; j < N; ++j)
         {
             size_t offset = j * lda + i_batch * stride;
@@ -611,7 +522,9 @@ inline void hipblas_init_sin<hipblasBfloat16>(
     hipblasBfloat16* A, size_t M, size_t N, size_t lda, size_t stride, size_t batch_count)
 {
     for(size_t i_batch = 0; i_batch < batch_count; i_batch++)
+#ifdef _OPENMP
 #pragma omp parallel for
+#endif
         for(size_t j = 0; j < N; ++j)
         {
             size_t offset = j * lda + i_batch * stride;
@@ -634,10 +547,10 @@ inline void hipblas_init_sin<hipblasBfloat16>(std::vector<hipblasBfloat16>& A,
 /*! \brief  symmetric matrix initialization: */
 // for real matrix only
 template <typename T>
-void hipblas_init_symmetric(std::vector<T>& A, int N, int lda)
+void hipblas_init_symmetric(std::vector<T>& A, int64_t N, int64_t lda)
 {
-    for(int i = 0; i < N; ++i)
-        for(int j = 0; j <= i; ++j)
+    for(int64_t i = 0; i < N; ++i)
+        for(int64_t j = 0; j <= i; ++j)
         {
             auto r         = random_generator<T>();
             A[j + i * lda] = r;
@@ -648,11 +561,11 @@ void hipblas_init_symmetric(std::vector<T>& A, int N, int lda)
 /*! \brief symmetric matrix initialization for strided_batched matricies: */
 template <typename T>
 void hipblas_init_symmetric(
-    std::vector<T>& A, int N, int lda, hipblasStride strideA, int batch_count)
+    std::vector<T>& A, int64_t N, int64_t lda, hipblasStride strideA, int64_t batch_count)
 {
-    for(int b = 0; b < batch_count; b++)
-        for(int off = b * strideA, i = 0; i < N; ++i)
-            for(int j = 0; j <= i; ++j)
+    for(int64_t b = 0; b < batch_count; b++)
+        for(int64_t off = b * strideA, i = 0; i < N; ++i)
+            for(int64_t j = 0; j <= i; ++j)
             {
                 auto r               = random_generator<T>();
                 A[i + j * lda + off] = r;
@@ -664,10 +577,10 @@ void hipblas_init_symmetric(
 // for complex matrix only, the real/imag part would be initialized with the same value
 // except the diagonal elment must be real
 template <typename T>
-void hipblas_init_hermitian(std::vector<T>& A, int N, int lda)
+void hipblas_init_hermitian(std::vector<T>& A, int64_t N, int64_t lda)
 {
-    for(int i = 0; i < N; ++i)
-        for(int j = 0; j <= i; ++j)
+    for(int64_t i = 0; i < N; ++i)
+        for(int64_t j = 0; j <= i; ++j)
             if(i == j)
                 A[j + i * lda] = random_generator<real_t<T>>();
             else
@@ -697,33 +610,34 @@ inline void hipblass_init_nan(
 /* ============================================================================================= */
 /*! \brief For testing purposes, to convert a regular matrix to a packed matrix.                  */
 template <typename T>
-inline void regular_to_packed(bool upper, const T* A, T* AP, int n)
+inline void regular_to_packed(bool upper, const T* A, T* AP, int64_t n)
 {
-    int index = 0;
+    int64_t index = 0;
     if(upper)
-        for(int i = 0; i < n; i++)
-            for(int j = 0; j <= i; j++)
+        for(int64_t i = 0; i < n; i++)
+            for(int64_t j = 0; j <= i; j++)
                 AP[index++] = A[j + i * n];
     else
-        for(int i = 0; i < n; i++)
-            for(int j = i; j < n; j++)
+        for(int64_t i = 0; i < n; i++)
+            for(int64_t j = i; j < n; j++)
                 AP[index++] = A[j + i * n];
 }
 
 /* ============================================================================ */
 /* \brief For testing purposes, to convert a regular matrix to a banded matrix. */
 template <typename T>
-inline void regular_to_banded(bool upper, const T* A, int lda, T* AB, int ldab, int n, int k)
+inline void regular_to_banded(
+    bool upper, const T* A, int64_t lda, T* AB, int64_t ldab, int64_t n, int64_t k)
 {
     // convert regular hA matrix to banded hAB matrix.
-    for(int j = 0; j < n; j++)
+    for(int64_t j = 0; j < n; j++)
     {
-        int min1 = upper ? std::max(0, j - k) : j;
-        int max1 = upper ? j : std::min(n - 1, j + k);
-        int m    = upper ? k - j : -j;
+        int64_t min1 = upper ? std::max(int64_t(0), j - k) : j;
+        int64_t max1 = upper ? j : std::min(n - 1, j + k);
+        int64_t m    = upper ? k - j : -j;
 
         // Move bands of hA into new banded hAB format.
-        for(int i = min1; i <= max1; i++)
+        for(int64_t i = min1; i <= max1; i++)
             AB[j * ldab + (m + i)] = A[j * lda + i];
 
         min1 = upper ? k + 1 : std::min(k + 1, n - j);
@@ -731,14 +645,14 @@ inline void regular_to_banded(bool upper, const T* A, int lda, T* AB, int ldab,
 
         // fill in bottom with random data to ensure we aren't using it.
         // for !upper, fill in bottom right triangle as well.
-        for(int i = min1; i <= max1; i++)
+        for(int64_t i = min1; i <= max1; i++)
             hipblas_init<T>(AB + j * ldab + i, 1, 1, 1);
 
         // for upper, fill in top left triangle with random data to ensure
         // we aren't using it.
         if(upper)
         {
-            for(int i = 0; i < m; i++)
+            for(int64_t i = 0; i < m; i++)
                 hipblas_init<T>(AB + j * ldab + i, 1, 1, 1);
         }
     }
@@ -747,12 +661,12 @@ inline void regular_to_banded(bool upper, const T* A, int lda, T* AB, int ldab,
 /* ============================================================================== */
 /* \brief For testing purposes, zeros out elements not needed in a banded matrix. */
 template <typename T>
-inline void banded_matrix_setup(bool upper, T* A, int lda, int n, int k)
+inline void banded_matrix_setup(bool upper, T* A, int64_t lda, int64_t n, int64_t k)
 {
     // Make A a banded matrix with k sub/super diagonals.
-    for(int i = 0; i < n; i++)
+    for(int64_t i = 0; i < n; i++)
     {
-        for(int j = 0; j < n; j++)
+        for(int64_t j = 0; j < n; j++)
         {
             if(upper && (j > k + i || i > j))
                 A[j * n + i] = T(0.0);
@@ -766,29 +680,29 @@ inline void banded_matrix_setup(bool upper, T* A, int lda, int n, int k)
 /*! \brief For testing purposes, makes a matrix hA into a unit_diagonal matrix and               *
  *         randomly initialize the diagonal.                                                     */
 template <typename T>
-void make_unit_diagonal(hipblasFillMode_t uplo, T* hA, int lda, int N)
+void make_unit_diagonal(hipblasFillMode_t uplo, T* hA, int64_t lda, int64_t N)
 {
     if(uplo == HIPBLAS_FILL_MODE_LOWER)
     {
-        for(int i = 0; i < N; i++)
+        for(int64_t i = 0; i < N; i++)
         {
-            T diag = hA[i + i * N];
-            for(int j = 0; j <= i; j++)
+            T diag = hA[i + i * lda];
+            for(int64_t j = 0; j <= i; j++)
                 hA[i + j * lda] = hA[i + j * lda] / diag;
         }
     }
     else // rocblas_fill_upper
     {
-        for(int j = 0; j < N; j++)
+        for(int64_t j = 0; j < N; j++)
         {
             T diag = hA[j + j * lda];
-            for(int i = 0; i <= j; i++)
+            for(int64_t i = 0; i <= j; i++)
                 hA[i + j * lda] = hA[i + j * lda] / diag;
         }
     }
 
     // randomly initalize diagonal to ensure we aren't using it's values for tests.
-    for(int i = 0; i < N; i++)
+    for(int64_t i = 0; i < N; i++)
     {
         hipblas_init<T>(hA + i * lda + i, 1, 1, 1);
     }
@@ -798,26 +712,28 @@ void make_unit_diagonal(hipblasFillMode_t uplo, T* hA, int lda, int N)
 /*! \brief For testing purposes, prepares matrix hA for a triangular solve.                      *
  *         Makes hA strictly diagonal dominant (SPD), then calculates Cholesky factorization     *
  *         of hA.                                                                                */
+/*
 template <typename T>
-void prepare_triangular_solve(T* hA, int lda, T* AAT, int N, char char_uplo)
+void prepare_triangular_solve(T* hA, int64_t lda, T* AAT, int64_t N, char char_uplo)
 {
     //  calculate AAT = hA * hA ^ T
-    cblas_gemm<T>(HIPBLAS_OP_N, HIPBLAS_OP_C, N, N, N, T(1.0), hA, lda, hA, lda, T(0.0), AAT, lda);
+    ref_gemm<T>(HIPBLAS_OP_N, HIPBLAS_OP_C, N, N, N, T(1.0), hA, lda, hA, lda, T(0.0), AAT, lda);
 
     //  copy AAT into hA, make hA strictly diagonal dominant, and therefore SPD
-    for(int i = 0; i < N; i++)
+    for(int64_t i = 0; i < N; i++)
     {
         T t = 0.0;
-        for(int j = 0; j < N; j++)
+        for(int64_t j = 0; j < N; j++)
         {
             hA[i + j * lda] = AAT[i + j * lda];
-            t += std::abs(AAT[i + j * lda]);
+            t += hipblas_abs(AAT[i + j * lda]);
         }
         hA[i + i * lda] = t;
     }
     //  calculate Cholesky factorization of SPD matrix hA
-    cblas_potrf<T>(char_uplo, N, hA, lda);
+    ref_potrf<T>(char_uplo, N, hA, lda);
 }
+*/
 
 /* ============================================================================================ */
 /*! \brief  turn float -> 's', double -> 'd', hipblas_float_complex -> 'c', hipblas_double_complex
@@ -834,12 +750,15 @@ int type2int(T val);
 /* ============================================================================================ */
 /*! \brief  Debugging purpose, print out CPU and GPU result matrix, not valid in complex number  */
 template <typename T, std::enable_if_t<!is_complex<T>, int> = 0>
-void print_matrix(
-    const std::vector<T>& CPU_result, const std::vector<T>& GPU_result, int m, int n, int lda)
+void print_matrix(const std::vector<T>& CPU_result,
+                  const std::vector<T>& GPU_result,
+                  int64_t               m,
+                  int64_t               n,
+                  int64_t               lda)
 {
-    for(int i = 0; i < m; i++)
-        for(int j = 0; j < n; j++)
-            printf("matrix  col %d, row %d, CPU result=%.8g, GPU result=%.8g\n",
+    for(int64_t i = 0; i < m; i++)
+        for(int64_t j = 0; j < n; j++)
+            printf("matrix  col %ld, row %ld, CPU result=%.8g, GPU result=%.8g\n",
                    i,
                    j,
                    double(CPU_result[j + i * lda]),
@@ -848,12 +767,15 @@ void print_matrix(
 
 /*! \brief  Debugging purpose, print out CPU and GPU result matrix, valid for complex number  */
 template <typename T, std::enable_if_t<+is_complex<T>, int> = 0>
-void print_matrix(
-    const std::vector<T>& CPU_result, const std::vector<T>& GPU_result, int m, int n, int lda)
+void print_matrix(const std::vector<T>& CPU_result,
+                  const std::vector<T>& GPU_result,
+                  int64_t               m,
+                  int64_t               n,
+                  int64_t               lda)
 {
-    for(int i = 0; i < m; i++)
-        for(int j = 0; j < n; j++)
-            printf("matrix  col %d, row %d, CPU result=(%.8g,%.8g), GPU result=(%.8g,%.8g)\n",
+    for(int64_t i = 0; i < m; i++)
+        for(int64_t j = 0; j < n; j++)
+            printf("matrix  col %ld, row %ld, CPU result=(%.8g,%.8g), GPU result=(%.8g,%.8g)\n",
                    i,
                    j,
                    double(CPU_result[j + i * lda].real()),
@@ -862,6 +784,44 @@ void print_matrix(
                    double(GPU_result[j + i * lda].imag()));
 }
 
+/* ============================================================================================= */
+/*! \brief For testing purposes, copy one matrix into another with different leading dimensions  */
+template <typename T, typename U>
+void copy_matrix_with_different_leading_dimensions(T&      hB,
+                                                   U&      hC,
+                                                   int64_t M,
+                                                   int64_t N,
+                                                   size_t  ldb,
+                                                   size_t  ldc,
+                                                   size_t  strideb     = 0,
+                                                   size_t  stridec     = 0,
+                                                   int64_t batch_count = 1)
+{
+    for(int64_t b = 0; b < batch_count; b++)
+    {
+        auto* B = hB + b * strideb;
+        auto* C = hC + b * stridec;
+        for(int64_t i = 0; i < M; i++)
+            for(int64_t j = 0; j < N; j++)
+                C[i + j * ldc] = B[i + j * ldb];
+    }
+}
+
+template <typename T, typename U>
+void copy_matrix_with_different_leading_dimensions_batched(
+    T& hB, U& hC, int64_t M, int64_t N, size_t ldb, size_t ldc)
+{
+    int64_t batch_count = hB.batch_count();
+    for(int64_t b = 0; b < batch_count; b++)
+    {
+        auto* B = hB[b];
+        auto* C = hC[b];
+        for(int64_t i = 0; i < M; i++)
+            for(int64_t j = 0; j < N; j++)
+                C[i + j * ldc] = B[i + j * ldb];
+    }
+}
+
 /* =============================================================================================== */
 
 /* ============================================================================================ */
@@ -872,6 +832,8 @@ std::string hipblas_exepath();
 // Temp directory rooted random path
 std::string hipblas_tempname();
 
+std::string getArchString();
+
 #endif // __cplusplus
 
 #ifdef __cplusplus
@@ -885,9 +847,34 @@ int query_device_property();
 /*  set current device to device_id */
 void set_device(int device_id);
 
+typedef enum hipblasClientProcessor
+{
+    // matching enum used in hipGcnArch
+    // only including supported types
+    gfx803  = 803,
+    gfx900  = 900,
+    gfx906  = 906,
+    gfx908  = 908,
+    gfx90a  = 910,
+    gfx940  = 940,
+    gfx941  = 941,
+    gfx942  = 942,
+    gfx1010 = 1010,
+    gfx1011 = 1011,
+    gfx1012 = 1012,
+    gfx1030 = 1030,
+    gfx1031 = 1031,
+    gfx1032 = 1032,
+    gfx1034 = 1034,
+    gfx1035 = 1035,
+    gfx1100 = 1100,
+    gfx1101 = 1101,
+    gfx1102 = 1102
+} hipblasClientProcessor;
+
 /* get architecture number */
-int getArch();
-int getArchMajor();
+hipblasClientProcessor getArch();
+int                    getArchMajor();
 
 /* ============================================================================================ */
 /*  timing: HIP only provides very limited timers function clock() and not general;
@@ -939,6 +926,10 @@ public:
     }
 };
 
+hipblasStatus_t hipblas_internal_convert_hip_to_hipblas_status(hipError_t status);
+
+hipblasStatus_t hipblas_internal_convert_hip_to_hipblas_status_and_log(hipError_t status);
+
 #include "hipblas_arguments.hpp"
 
 #endif // __cplusplus
